import difflib
import typing
import numpy as np
import pandas as pd
from box import Box, box_list
import plotly.graph_objects as go  # type:ignore
from typing import Union, Literal, get_args, Any, cast


def find_closest_match(given_str: str, available_str: list[str]):
    closest_match = difflib.get_close_matches(given_str, available_str, n=1)
    if closest_match:
        return closest_match[0]
    else:
        return None


DCType = Literal[
    "DataFrame",
    "Grayscale",
    "Image",
    "Matrix",
    "OrderedPair",
    "OrderedTriple",
    "Plotly",
    "Bytes",
    "TextBlob",
    "Scalar",
    "Surface",
    "Vector",
    "ParametricDataFrame",
    "ParametricGrayscale",
    "ParametricImage",
    "ParametricMatrix",
    "ParametricOrderedPair",
    "ParametricOrderedTriple",
    "ParametricPlotly",
    "ParametricScalar",
    "ParametricSurface",
    "ParametricVector",
]

DCNpArrayType = np.ndarray[Union[int, float], np.dtype[Any]]

DCKwargsValue = Union[
    list[Union[int, float]],
    int,
    float,
    dict[str, Union[int, float, DCNpArrayType]],
    DCNpArrayType,
    pd.DataFrame,
    go.Figure,
    bytes,
    str,
    None,
]

ExtraType = dict[str, Any] | None


class DataContainer(Box):
    """
    A class that processes various types of data and supports dot assignment

    Learn more: https://github.com/flojoy-io/flojoy-python/issues/4

    Usage
    -----
    import numpy as np

    v = DataContainer()

    v.x = np.linspace(1,20,0.1)

    v.y = np.sin(v.x)

    v.type = 'OrderedPair'

    """

    allowed_types = list(typing.get_args(DCType))
    allowed_keys = [
        "x",
        "y",
        "z",
        "t",
        "v",
        "m",
        "c",
        "r",
        "g",
        "b",
        "a",
        "text_blob",
        "fig",
        "extra",
    ]
    combinations = {
        "x": ["y", "t", "z", "extra"],
        "y": ["x", "t", "z", "extra"],
        "z": ["x", "y", "t", "extra"],
        "c": ["t", "extra"],
        "v": ["t", "extra"],
        "m": ["t", "extra"],
        "t": [*(value for value in allowed_keys if value not in ["t"])],
        "r": ["g", "b", "t", "a", "extra"],
        "g": ["r", "b", "t", "a", "extra"],
        "b": ["r", "g", "t", "a", "extra"],
        "a": ["r", "g", "b", "t", "extra"],
        "bytes": ["extra"],
        "text_blob": ["extra"],
        "extra": [*(k for k in allowed_keys if k not in ["extra"])],
        "fig": ["t", "extra"],
    }
    type_keys_map: dict[DCType, list[str]] = {
        "DataFrame": ["m"],
        "Matrix": ["m"],
        "Vector": ["v"],
        "Grayscale": ["m"],
        "Image": ["r", "g", "b", "a"],
        "OrderedPair": ["x", "y"],
        "OrderedTriple": ["x", "y", "z"],
        "Surface": ["x", "y", "z"],
        "Scalar": ["c"],
        "Plotly": ["fig"],
        "Bytes": ["b"],
        "TextBlob": ["text_blob"],
    }

    SKIP_ARRAYIEFY_TYPES = [
        str,
        bytes,
        go.Figure,
        pd.DataFrame,
        np.ndarray,
    ]  # value types not to be arrayified

    type: DCType

    def copy(self):
        # Create an instance of DataContainer class
        copied_instance = DataContainer(**self)
        return copied_instance

    def _ndarrayify(
        self, value: DCKwargsValue
    ) -> Union[DCNpArrayType, pd.DataFrame, dict[str, DCNpArrayType], go.Figure, None]:
        if isinstance(value, int) or isinstance(value, float):
            return np.array([value])
        elif isinstance(value, dict):
            arrayified_value: dict[str, DCNpArrayType] = {}
            for k, v in value.items():
                arrayified_value[k] = cast(DCNpArrayType, self._ndarrayify(v))
            return arrayified_value
        elif isinstance(value, box_list.BoxList):
            arrayified_value: dict[str, DCNpArrayType] = {}
            for k, v in value.__dict__.items():
                arrayified_value[k] = cast(DCNpArrayType, self._ndarrayify(v))
            return arrayified_value
        elif isinstance(value, list):
            return np.array(value)
        elif value is None:
            return value
        else:
            raise ValueError(
                f"DataContainer keys must be any of "
                f"following types: {get_args(DCKwargsValue)}"
            )

    def __init__(  # type:ignore
        self, type: DCType = "OrderedPair", **kwargs: DCKwargsValue
    ):
        self.type = type
        for k, v in kwargs.items():
            self[k] = v

    def __getattribute__(self, __name: str) -> Any:
        return super().__getattribute__(__name)

    def __getitem__(self, key: str, _ignore_default: bool = False) -> Any:
        return super().__getitem__(key, _ignore_default)  # type:ignore

    def __setitem__(self, key: str, value: DCKwargsValue) -> None:
        if (
            key not in ["type", "extra", "c"]
            and type(value) not in self.SKIP_ARRAYIEFY_TYPES
        ):
            formatted_value = self._ndarrayify(value)
            super().__setitem__(key, formatted_value)  # type:ignore
        else:
            super().__setitem__(key, value)  # type: ignore

    def __check_combination(self, key: str, keys: list[str], allowed_keys: list[str]):
        for i in keys:
            if i not in allowed_keys:
                raise ValueError(
                    f"You can't have '{key}' and '{i}' keys together for '{self.type}' type!"
                )

    def __validate_key_for_type(self, data_type: DCType, key: str):
        if data_type.startswith("parametric_") and key != "t":
            splitted_type = cast(DCType, data_type.split("parametric_")[1])
            self.__validate_key_for_type(splitted_type, key)
        else:
            if (
                key not in self.type_keys_map[data_type] + ["extra"]
                and data_type != "plotly"
            ):
                raise KeyError(
                    self.__build_error_text(
                        key, data_type, self.type_keys_map[data_type]
                    )
                )

    def __check_for_missing_keys(self, dc_type: DCType, keys: list[str]):
        if dc_type.startswith("parametric_"):
            if "t" not in keys:
                raise KeyError(f't key must be provided for "{dc_type}"')
            t = self["t"]
            is_ascending_order = all(t[i] <= t[i + 1] for i in range(len(t) - 1))
            if is_ascending_order is not True:
                raise ValueError("t key must be in ascending order")
            splitted_type = cast(DCType, dc_type.split("parametric_")[1])
            self.__check_for_missing_keys(splitted_type, keys)
        else:
            for k in self.type_keys_map[dc_type]:
                if k not in keys:
                    raise KeyError(f'"{k}" key must be provided for type "{dc_type}"')

    def __build_error_text(self, key: str, data_type: str, available_keys: list[str]):
        return (
            f'Invalid key "{key}" provided for data type "{data_type}", '
            f'supported keys: {", ".join(available_keys)}'
        )

    def validate(self):
        dc_type = self.type

        if dc_type not in self.allowed_types:
            closest_type = find_closest_match(dc_type, self.allowed_types)
            helper_text = (
                f'Did you mean: "{closest_type}" ?'
                if closest_type
                else f'allowed types: "{", ".join(self.allowed_types)}"'
            )
            raise ValueError(
                f'unsupported type "{dc_type}" passed to '
                f"DataContainer class, {helper_text}"
            )
        dc_keys = list(cast(list[str], self.keys()))
        for k in dc_keys:
            if k != "type":
                self.__check_combination(
                    k,
                    list(key for key in dc_keys if key not in ["type", k]),
                    self.combinations[k],
                )
                self.__validate_key_for_type(dc_type, k)
        self.__check_for_missing_keys(dc_type, dc_keys)


class OrderedPair(DataContainer):
    x: DCNpArrayType
    y: DCNpArrayType

    def __init__(  # type:ignore
        self, x: DCNpArrayType, y: DCNpArrayType, extra: ExtraType = None
    ):
        super().__init__(type="OrderedPair", x=x, y=y, extra=extra)


class ParametricOrderedPair(DataContainer):
    x: DCNpArrayType
    y: DCNpArrayType
    t: DCNpArrayType

    def __init__(  # type:ignore
        self,
        x: DCNpArrayType,
        y: DCNpArrayType,
        t: DCNpArrayType,
        extra: ExtraType = None,
    ):
        super().__init__(type="ParametricOrderedPair", x=x, y=y, t=t, extra=extra)


class OrderedTriple(DataContainer):
    x: DCNpArrayType
    y: DCNpArrayType
    z: DCNpArrayType

    def __init__(  # type:ignore
        self,
        x: DCNpArrayType,
        y: DCNpArrayType,
        z: DCNpArrayType,
        extra: ExtraType = None,
    ):
        super().__init__(type="OrderedTriple", x=x, y=y, z=z, extra=extra)


class ParametricOrderedTriple(DataContainer):
    x: DCNpArrayType
    y: DCNpArrayType
    z: DCNpArrayType
    t: DCNpArrayType

    def __init__(  # type:ignore
        self,
        x: DCNpArrayType,
        y: DCNpArrayType,
        z: DCNpArrayType,
        t: DCNpArrayType,
        extra: ExtraType = None,
    ):
        super().__init__(
            type="ParametricOrderedTriple", x=x, y=y, z=z, t=t, extra=extra
        )


class Surface(DataContainer):
    x: DCNpArrayType
    y: DCNpArrayType
    z: DCNpArrayType

    def __init__(  # type:ignore
        self,
        x: DCNpArrayType,
        y: DCNpArrayType,
        z: DCNpArrayType,
        extra: ExtraType = None,
    ):
        super().__init__(type="Surface", x=x, y=y, z=z, extra=extra)

    def validate(self):
        if self.z.ndim < 2:
            raise ValueError("z key must be of 2D array for Surface type!")
        super().validate()


class ParametricSurface(DataContainer):
    x: DCNpArrayType
    y: DCNpArrayType
    z: DCNpArrayType
    t: DCNpArrayType

    def __init__(  # type:ignore
        self,
        x: DCNpArrayType,
        y: DCNpArrayType,
        z: DCNpArrayType,
        t: DCNpArrayType,
        extra: ExtraType = None,
    ):
        super().__init__(type="ParametricSurface", x=x, y=y, z=z, t=t, extra=extra)

    def validate(self):
        if self.z.ndim < 2:
            raise ValueError("z key must be of 2D array for Surface type!")
        super().validate()


class Scalar(DataContainer):
    c: int | float

    def __init__(self, c: int | float, extra: ExtraType = None):  # type:ignore
        super().__init__(type="Scalar", c=c, extra=extra)


class ParametricScalar(DataContainer):
    c: int | float
    t: DCNpArrayType

    def __init__(  # type: ignore
        self, c: int | float, t: DCNpArrayType, extra: ExtraType = None
    ):
        super().__init__(type="ParametricScalar", c=c, t=t, extra=extra)


class Vector(DataContainer):
    v: DCNpArrayType

    def __init__(self, v: DCNpArrayType, extra: ExtraType = None):  # type:ignore
        super().__init__(type="Vector", v=v, extra=extra)


class ParametricVector(DataContainer):
    v: DCNpArrayType

    def __init__(  # type: ignore
        self, v: DCNpArrayType, t: DCNpArrayType, extra: ExtraType = None
    ):
        super().__init__(type="ParametricVector", v=v, t=t, extra=extra)


class Matrix(DataContainer):
    m: DCNpArrayType

    def __init__(self, m: DCNpArrayType, extra: ExtraType = None):  # type:ignore
        super().__init__(type="Matrix", m=m, extra=extra)


class ParametricMatrix(DataContainer):
    m: DCNpArrayType
    t: DCNpArrayType

    def __init__(  # type: ignore
        self, m: DCNpArrayType, t: DCNpArrayType, extra: ExtraType = None
    ):
        super().__init__(type="ParametricMatrix", m=m, t=t, extra=extra)


class DataFrame(DataContainer):
    m: pd.DataFrame

    def __init__(self, df: pd.DataFrame, extra: ExtraType = None):  # type:ignore
        super().__init__(type="DataFrame", m=df, extra=extra)


class ParametricDataFrame(DataContainer):
    m: pd.DataFrame
    t: DCNpArrayType

    def __init__(  # type:ignore
        self, df: pd.DataFrame, t: DCNpArrayType, extra: ExtraType = None
    ):
        super().__init__(type="ParametricDataFrame", m=df, t=t, extra=extra)


class Plotly(DataContainer):
    fig: go.Figure

    def __init__(self, fig: go.Figure, extra: ExtraType = None):  # type:ignore
        super().__init__(type="Plotly", fig=fig, extra=extra)


class ParametricPlotly(DataContainer):
    fig: go.Figure
    t: DCNpArrayType

    def __init__(  # type:ignore
        self, fig: go.Figure, t: DCNpArrayType, extra: ExtraType = None
    ):
        super().__init__(type="ParametricPlotly", fig=fig, t=t, extra=extra)


class Image(DataContainer):
    r: DCNpArrayType
    g: DCNpArrayType
    b: DCNpArrayType
    a: DCNpArrayType | None

    def __init__(  # type:ignore
        self,
        r: DCNpArrayType,
        g: DCNpArrayType,
        b: DCNpArrayType,
        a: DCNpArrayType | None = None,
        extra: ExtraType = None,
    ):
        super().__init__(type="Image", r=r, g=g, b=b, a=a, extra=extra)


class Bytes(DataContainer):
    b: bytes

    def __init__(
        self,
        b: bytes,
    ):
        super().__init__(type="Bytes", b=b)


class TextBlob(DataContainer):
    text_blob: str

    def __init__(self, text_blob: str):
        super().__init__(type="TextBlob", text_blob=text_blob)


class ParametricImage(DataContainer):
    t: DCNpArrayType
    r: DCNpArrayType
    g: DCNpArrayType
    b: DCNpArrayType
    a: DCNpArrayType | None

    def __init__(  # type:ignore
        self,
        r: DCNpArrayType,
        g: DCNpArrayType,
        b: DCNpArrayType,
        a: DCNpArrayType,
        t: DCNpArrayType,
        extra: ExtraType = None,
    ):
        super().__init__(type="ParametricImage", r=r, g=g, b=b, a=a, t=t, extra=extra)


class Grayscale(DataContainer):
    m: DCNpArrayType

    def __init__(self, img: DCNpArrayType, extra: ExtraType = None):  # type:ignore
        super().__init__(type="Grayscale", m=img, extra=extra)


class ParametricGrayscale(DataContainer):
    m: DCNpArrayType
    t: DCNpArrayType

    def __init__(  # type:ignore
        self, img: DCNpArrayType, t: DCNpArrayType, extra: ExtraType = None
    ):
        super().__init__(type="ParametricGrayscale", m=img, t=t, extra=extra)



"""
Nodes should try to accomodate any reasonable combination of inputs that a first-time Flojoy Studio user might try.

For example, the ADD node should make a best effort to do something reasonable when a matrix is added to a DataFrame, or a 2 matrices of a different size are added.

For this reason, we've created the `Reconciler` class to handle the process of turning different data types into compatible, easily added objects. 
"""
from typing import Tuple
import numpy

from .data_container import DataContainer


class IrreconcilableContainersException(Exception):
    pass


class Reconciler:
    def __init__(self, pad: float = 0):
        self.pad = pad

    def reconcile(
        self, lhs: DataContainer, rhs: DataContainer
    ) -> Tuple[DataContainer, DataContainer]:
        types_to_reconcile = set([lhs.type, rhs.type])
        if types_to_reconcile == set(["Matrix"]):
            return self.reconcile_matrix(lhs, rhs)
        elif types_to_reconcile == set(["DataFrame"]):
            return self.reconcile_dataframe(lhs, rhs)
        elif types_to_reconcile == set(["OrderedPair"]):
            return self.reconcile_ordered_pair(lhs, rhs)
        elif types_to_reconcile == set(["Matrix", "Scalar"]):
            return self.reconcile_matrix_scalar(lhs, rhs)
        elif types_to_reconcile == set(["Matrix", "DataFrame"]):
            return self.reconcile_dataframe_matrix(lhs, rhs)
        elif types_to_reconcile == set(["Scalar", "DataFrame"]):
            return self.reconcile_dataframe_scalar(lhs, rhs)
        else:
            raise IrreconcilableContainersException(
                "FloJoy doesn't know how to reconcile data containers of type %s and %s"
                % (lhs.type, rhs.type)
            )

    def reconcile_matrix(
        self, lhs: DataContainer, rhs: DataContainer
    ) -> Tuple[DataContainer, DataContainer]:
        # make the matrices equal sizes, by padding
        final_r = max(lhs.m.shape[0], rhs.m.shape[0])
        final_c = max(lhs.m.shape[1], rhs.m.shape[1])

        new_lhs = numpy.pad(
            lhs.m,
            ((0, final_r - lhs.m.shape[0]), (0, final_c - lhs.m.shape[1])),
            "constant",
            constant_values=self.pad,
        )
        new_rhs = numpy.pad(
            rhs.m,
            ((0, final_r - rhs.m.shape[0]), (0, final_c - rhs.m.shape[1])),
            "constant",
            constant_values=self.pad,
        )

        return (
            DataContainer(type="Matrix", m=new_lhs),
            DataContainer(type="Matrix", m=new_rhs),
        )

    def reconcile_dataframe(
        self, lhs: DataContainer, rhs: DataContainer
    ) -> Tuple[DataContainer, DataContainer]:
        # pandas' handling for DataFrames is actually pretty permissive. Let's just
        #  return both types as normal
        return (lhs, rhs)

    def reconcile_dataframe_scalar(
        self, lhs: DataContainer, rhs: DataContainer
    ) -> Tuple[DataContainer, DataContainer]:
        # let's expand the scalar to be a DataFrame the same size as the other DataFrame
        if lhs.type == "DataFrame":
            new_m = lhs.m.copy()
            new_m.iloc[:] = rhs.c
            return lhs, DataContainer(type="DataFrame", m=new_m)

        new_m = rhs.m.copy()
        new_m.iloc[:] = lhs.c
        return DataContainer(type="DataFrame", m=new_m), rhs

    def reconcile_ordered_pair(
        self, lhs: DataContainer, rhs: DataContainer
    ) -> Tuple[DataContainer, DataContainer]:
        raise NotImplementedError("TODO")

    def reconcile_matrix_scalar(
        self, lhs: DataContainer, rhs: DataContainer
    ) -> Tuple[DataContainer, DataContainer]:
        raise NotImplementedError("TODO")

    def reconcile_dataframe_matrix(
        self, lhs: DataContainer, rhs: DataContainer
    ) -> Tuple[DataContainer, DataContainer]:
        raise NotImplementedError("TODO")



from PIL import Image as PILImage
import json
import requests
import pandas as pd
import numpy as np
from pydantic import validator
from typing import Optional, Generic, TypeVar
from pydantic.generics import GenericModel


from flojoy import DataContainer
from flojoy.data_container import (
    OrderedPair,
    OrderedTriple,
    DataFrame,
    Matrix,
    Grayscale,
    Vector,
    Scalar,
    Image,
)
from flojoy.utils import PlotlyJSONEncoder


class NumpyEncoder(json.JSONEncoder):
    """json encoder for numpy types."""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


DataC = TypeVar("DataC")


class DefaultModel(GenericModel, Generic[DataC]):
    ref: str
    dataContainer: dict
    workspaceId: str
    location: str
    note: str


class OrderedPairModel(DefaultModel[DataC], Generic[DataC]):
    data: Optional[DataC]

    @validator("dataContainer")
    def type_must_match(cls, dc):
        assert dc["type"] == "OrderedPair", "dataContainer type does not match."
        return dc

    @validator("dataContainer")
    def keys_must_match(cls, dc):
        assert "x" in dc, 'dataContainer does not contain "x" dataset.'
        assert "y" in dc, 'dataContainer does not contain "y" dataset.'
        assert isinstance(dc["x"], list), '"x" dataset is not a list'
        assert isinstance(dc["y"], list), '"y" dataset is not a list'
        return dc


class OrderedTripleModel(DefaultModel[DataC], Generic[DataC]):
    data: Optional[DataC]

    @validator("dataContainer")
    def type_must_match(cls, dc):
        assert dc["type"] == "OrderedTriple", "dataContainer type does not match."
        return dc

    @validator("dataContainer")
    def keys_must_match(cls, dc):
        assert "x" in dc, 'dataContainer does not contain "x" dataset.'
        assert "y" in dc, 'dataContainer does not contain "y" dataset.'
        assert "z" in dc, 'dataContainer does not contain "z" dataset.'
        assert isinstance(dc["x"], list), '"x" dataset is not a list'
        assert isinstance(dc["y"], list), '"y" dataset is not a list'
        assert isinstance(dc["z"], list), '"z" dataset is not a list'
        return dc


class DataFrameModel(DefaultModel[DataC], Generic[DataC]):
    data: Optional[DataC]

    @validator("dataContainer")
    def type_must_match(cls, dc):
        assert dc["type"] == "DataFrame", "dataContainer type does not match."
        return dc

    @validator("dataContainer")
    def keys_must_match(cls, dc):
        assert "m" in dc, 'dataContainer does not contain "m" dataset.'
        assert isinstance(
            dc["m"], dict
        ), f'"m" dataset is not a list, type: {type(dc["m"])}'
        return dc


class MatrixModel(DefaultModel[DataC], Generic[DataC]):
    data: Optional[DataC]

    @validator("dataContainer")
    def type_must_match(cls, dc):
        assert dc["type"] == "Matrix", "dataContainer type does not match."
        return dc

    @validator("dataContainer")
    def keys_must_match(cls, dc):
        assert "m" in dc, 'dataContainer does not contain "m" dataset.'
        assert isinstance(dc["m"], list), '"m" dataset is not a list'
        assert isinstance(dc["m"][0], list), '"m" dataset is not a matrix'
        assert not isinstance(dc["m"][0][0], list), '"m" dataset is not 2D'
        return dc


class GrayscaleModel(DefaultModel[DataC], Generic[DataC]):
    data: Optional[DataC]

    @validator("dataContainer")
    def type_must_match(cls, dc):
        assert dc["type"] == "Grayscale", "dataContainer type does not match."
        return dc

    @validator("dataContainer")
    def keys_must_match(cls, dc):
        assert "m" in dc, 'dataContainer does not contain "m" dataset.'
        assert isinstance(dc["m"], list), '"m" dataset is not a list'
        assert isinstance(dc["m"][0], list), '"m" dataset is not a matrix'
        assert not isinstance(dc["m"][0][0], list), '"m" dataset is not 2D'
        return dc


class ScalarModel(DefaultModel[DataC], Generic[DataC]):
    data: Optional[DataC]

    @validator("dataContainer")
    def type_must_match(cls, dc):
        assert dc["type"] == "Scalar", "dataContainer type does not match."
        return dc

    @validator("dataContainer")
    def keys_must_match(cls, dc):
        assert "c" in dc, 'dataContainer does not contain "c" dataset.'
        assert isinstance(dc["c"], float) or isinstance(
            dc["c"], int
        ), '"c" dataset is not a float or int'
        return dc


class VectorModel(DefaultModel[DataC], Generic[DataC]):
    data: Optional[DataC]

    @validator("dataContainer")
    def type_must_match(cls, dc):
        assert dc["type"] == "Vector", "dataContainer type does not match."
        return dc

    @validator("dataContainer")
    def keys_must_match(cls, dc):
        assert "v" in dc, 'dataContainer does not contain "v" dataset.'
        assert isinstance(dc["v"], list)
        return dc


class ImageModel(DefaultModel[DataC], Generic[DataC]):
    data: Optional[DataC]

    @validator("dataContainer")
    def type_must_match(cls, dc):
        assert dc["type"] == "Image", "dataContainer type does not match."
        return dc

    @validator("dataContainer")
    def keys_must_match(cls, dc):
        assert "r" in dc, 'dataContainer does not contain "r" dataset.'
        assert "g" in dc, 'dataContainer does not contain "g" dataset.'
        assert "b" in dc, 'dataContainer does not contain "b" dataset.'
        assert isinstance(dc["r"], list), '"r" dataset is not a list'
        assert isinstance(dc["g"], list), '"g" dataset is not a list'
        assert isinstance(dc["b"], list), '"b" dataset is not a list'
        return dc


def check_deserialize(response):
    dc_type = response["dataContainer"]["type"]
    match dc_type:
        case "OrderedPair":
            OrderedPairModel.parse_obj(response)
        case "OrderedTriple":
            OrderedTripleModel.parse_obj(response)
        case "DataFrame":
            DataFrameModel.parse_obj(response)
        case "Matrix":
            MatrixModel.parse_obj(response)
        case "Grayscale":
            GrayscaleModel.parse_obj(response)
        case "Scalar":
            ScalarModel.parse_obj(response)
        case "Vector":
            VectorModel.parse_obj(response)
        case "Image":
            ImageModel.parse_obj(response)
        case _:
            raise TypeError(
                f"Unsupported DataContainer type: {dc_type}. Check case (e.g. OrderedPair)."
            )


class FlojoyCloud:
    """
    A class that allows pulling and pushing DataContainers from the
    Flojoy cloud client (cloud.flojoy.ai).

    Returns data in a pythonic format (e.g. Pillow for images,
    DataFrames for arrays/matrices).

    Will support the majority of the Flojoy cloud API:
    https://rest.flojoy.ai/api-reference

    Recommended for api key:
    utils.get_credentials()[0]["value"]
    or
    os.environ.get("FLOJOY_CLOUD_KEY")
    """

    def __init__(self, api_key: str):
        self.headers = {"api_key": api_key}
        self.base_url = "https://cloud.flojoy.ai/api/v1"
        self.valid_types = [
            "OrderedPair",
            "OrderedTriple",
            "DataFrame",
            "Grayscale",
            "Matrix",
            "Scalar",
            "Vector",
            "Image",
        ]

    def _create_payload(self, data, dc_type: str) -> str:
        """
        A method that formats data into a payload that can be handled by
        the Flojoy cloud client.
        """
        if isinstance(data, DataContainer):
            return json.dumps({"data": data}, cls=PlotlyJSONEncoder)

        assert (
            dc_type in self.valid_types
        ), f"Type {dc_type} not supported. Check capitals (e.g. OrderedPair)."
        match dc_type:
            case "OrderedPair":
                if not (isinstance(data, dict) and "x" in data and "y" in data):
                    raise TypeError(
                        "For ordered pair type, data must be in dictionary form with keys 'x' and 'y'"
                    )
                payload = json.dumps(
                    {
                        "data": {
                            "type": "OrderedPair",
                            "x": data["x"],
                            "y": data["y"],
                        }
                    },
                    cls=NumpyEncoder,
                )
        match dc_type:
            case "OrderedTriple":
                if isinstance(data, dict) and "x" in data:
                    payload = json.dumps(
                        {
                            "data": {
                                "type": "OrderedTriple",
                                "x": data["x"],
                                "y": data["y"],
                                "z": data["z"],
                            }
                        },
                        cls=NumpyEncoder,
                    )
                else:
                    print(
                        "For ordered triple type, data must be in"
                        " dictionary form with keys 'x', 'y', and 'z'"
                    )
                    raise TypeError
            case "DataFrame":
                payload = json.dumps(
                    {"data": {"type": "DataFrame", "m": data}}, cls=PlotlyJSONEncoder
                )
            case "Matrix":
                payload = json.dumps(
                    {"data": {"type": "Matrix", "m": data}}, cls=NumpyEncoder
                )
            case "Grayscale":
                payload = json.dumps(
                    {"data": {"type": "Grayscale", "m": data}}, cls=NumpyEncoder
                )
            case "Scalar":
                payload = json.dumps(
                    {"data": {"type": "Scalar", "c": data}}, cls=NumpyEncoder
                )
            case "Vector":
                payload = json.dumps(
                    {"data": {"type": "Vector", "v": data}}, cls=NumpyEncoder
                )
            case "Image":
                RGB_img = np.asarray(data)
                red_channel = RGB_img[:, :, 0]
                green_channel = RGB_img[:, :, 1]
                blue_channel = RGB_img[:, :, 2]

                if RGB_img.shape[2] == 4:
                    alpha_channel = RGB_img[:, :, 3]
                else:
                    alpha_channel = None
                payload = json.dumps(
                    {
                        "data": {
                            "type": "Image",
                            "r": red_channel,
                            "g": green_channel,
                            "b": blue_channel,
                            "a": alpha_channel,
                        }
                    },
                    cls=NumpyEncoder,
                )

        return payload

    def fetch_dc(self, dc_id: str) -> dict:
        """
        A method that retrieves DataContainers from the Flojoy cloud.
        """
        url = f"{self.base_url}/dcs/{dc_id}"
        response = requests.request("GET", url, headers=self.headers)
        response = json.loads(response.text)
        check_deserialize(response)
        return response

    def to_python(self, dc: dict) -> pd.DataFrame | float | list | PILImage.Image:
        """
        A method that converts data from DataContainers into pythonic
        data types like Pillow for images.
        """
        dc_type = dc["dataContainer"]["type"]
        match dc_type:
            case "OrderedPair" | "OrderedTriple":
                df = pd.DataFrame(dc["dataContainer"])
                return df.drop(columns=["type"])
            case "DataFrame" | "Matrix" | "Grayscale":
                df = pd.DataFrame(dc["dataContainer"]["m"])
                return df
            case "Scalar":
                return float(dc["dataContainer"]["c"])
            case "Vector":
                return list(dc["dataContainer"]["v"])
            case "Image":
                image = dc["dataContainer"]
                r = image["r"]
                g = image["g"]
                b = image["b"]
                if "a" in image:
                    a = image["a"]
                    img_combined = np.stack((r, g, b, a), axis=2)
                    return PILImage.fromarray(np.uint8(img_combined)).convert("RGBA")
                else:
                    img_combined = np.stack((r, g, b), axis=2)
                    return PILImage.fromarray(np.uint8(img_combined)).convert("RGB")

    def to_dc(self, dc: dict) -> DataContainer:
        """
        A method that converts data from DataContainers into pythonic
        data types like Pillow for images.
        """
        dc = dc["dataContainer"]
        match dc["type"]:
            case "OrderedPair":
                return OrderedPair(x=np.array(dc["x"]), y=np.array(dc["y"]))
            case "OrderedTriple":
                return OrderedTriple(
                    x=np.array(dc["x"]), y=np.array(dc["y"]), z=np.array(dc["z"])
                )
            case "DataFrame":
                return DataFrame(df=pd.DataFrame(dc["m"]))
            case "Matrix":
                return Matrix(m=np.array(dc["m"]))
            case "Grayscale":
                return Grayscale(img=np.array(dc["m"]))
            case "Scalar":
                return Scalar(c=float(dc["c"]))
            case "Vector":
                return Vector(v=np.array(dc["v"]))
            case "Image":
                r = np.array(dc["r"], dtype=np.uint8)
                g = np.array(dc["g"], dtype=np.uint8)
                b = np.array(dc["b"], dtype=np.uint8)
                if "a" in dc:
                    a = np.array(dc["a"], dtype=np.uint8)
                    return Image(r=r, g=g, b=b, a=a)
                else:
                    return Image(r=r, g=g, b=b)
            case _:
                raise Exception("Unknown data container type")

    def create_measurement(self, name: str, privacy: str = "private") -> dict:
        """
        A method that creates a measurements with the name specified.
        """
        url = f"{self.base_url}/measurements"
        payload = json.dumps({"name": name, "privacy": privacy})
        response = requests.request("POST", url, headers=self.headers, data=payload)
        response = json.loads(response.text)

        return response

    def list_measurements(self, size: int = 10) -> list:
        """
        A method that lists the number of measurements specified.

        If the number of measurements is less than the specified number,
        an error will be thrown.
        """
        url = f"{self.base_url}/measurements/?size={size}"
        response = requests.request("GET", url, headers=self.headers)
        response = json.loads(response.text)
        response = response["data"]

        return response

    def fetch_measurement(self, meas_id: str):
        """
        A method fetchs measurements from the client.
        """
        url = f"{self.base_url}/measurements/{meas_id}"
        response = requests.request("GET", url, headers=self.headers)
        response = json.loads(response.text)

        return response

    def rename_measurement(self, meas_id: str, name: str):
        """
        Rename the specified measurement.
        """
        url = f"{self.base_url}/measurements/{meas_id}"
        payload = payload = json.dumps({"name": name})
        response = requests.request("PATCH", url, headers=self.headers, data=payload)
        response = json.loads(response.text)

        return response

    def store_dc(self, data, dc_type: str, meas_id: str):
        """
        A method that stores a formatted data payload in a measurement.
        """
        url = f"{self.base_url}/dcs/add/{meas_id}"
        payload = self._create_payload(data, dc_type)
        response = requests.request("POST", url, headers=self.headers, data=payload)

        return json.loads(response.text)



import json
import traceback
from functools import wraps

from flojoy.node_init import NodeInitService
from .data_container import DataContainer
from .utils import PlotlyJSONEncoder
from typing import Callable, Any, Optional
from .job_result_utils import get_frontend_res_obj_from_result, get_dc_from_result
from .utils import send_to_socket
from .config import logger
from .parameter_types import format_param_value
from inspect import signature
from .job_service import JobService

__all__ = ["flojoy", "DefaultParams"]


def fetch_inputs(previous_jobs: list[dict[str, str]]):
    """
    Queries for job results

    Parameters
    ----------
    previous_jobs : list of jobs that directly precede this node.
    Each item representing a job contains `job_id` and `input_name`.
    `input_name` is the port where the previous job with `job_id` connects to.

    Returns
    -------
    inputs : list of DataContainer objects
    """
    dict_inputs: dict[str, DataContainer | list[DataContainer]] = dict()

    try:
        for prev_job in previous_jobs:
            prev_job_id = prev_job.get("job_id")
            input_name = prev_job.get("input_name", "")
            multiple = prev_job.get("multiple", False)
            edge = prev_job.get("edge", "")

            logger.debug(
                "fetching input from prev job id:",
                prev_job_id,
                " for input:",
                input_name,
                "edge: ",
                edge,
            )

            job_result = JobService().get_job_result(prev_job_id)
            if not job_result:
                raise ValueError(
                    f"Tried to get job result from {prev_job_id} but it was None"
                )

            result = (
                get_dc_from_result(job_result[edge])
                if edge != "default"
                else get_dc_from_result(job_result)
            )
            if result is not None:
                logger.debug(f"got job result from {prev_job_id}")
                if multiple:
                    if input_name not in dict_inputs:
                        dict_inputs[input_name] = [result]
                    else:
                        dict_inputs[input_name].append(result)
                else:
                    dict_inputs[input_name] = result

    except Exception:
        logger.debug(traceback.format_exc())

    return dict_inputs


class DefaultParams:
    def __init__(
        self, node_id: str, job_id: str, jobset_id: str, node_type: str
    ) -> None:
        self.node_id = node_id
        self.job_id = job_id
        self.jobset_id = jobset_id
        self.node_type = node_type


def flojoy(
    original_function: Callable[..., Optional[DataContainer | dict[str, Any]]]
    | None = None,
    *,
    node_type: Optional[str] = None,
    deps: Optional[dict[str, str]] = None,
    inject_node_metadata: bool = False,
):
    """
    Decorator to turn Python functions with numerical return
    values into Flojoy nodes.

    @flojoy is intended to eliminate boilerplate in connecting
    Python scripts as visual nodes

    Into whatever function it wraps, @flojoy injects
    1. the last node's input as list of DataContainer object
    2. parameters for that function (either set by the user or default)

    Parameters
    ----------
    `func`: Python function that returns DataContainer object

    Returns
    -------
    A dict containing DataContainer object

    Usage Example
    -------------
    ```
    @flojoy
    def SINE(dc_inputs:list[DataContainer], params:dict[str, Any]):

        logger.debug('params passed to SINE', params)

        dc_input = dc_inputs[0]

        output = DataContainer(
            x=dc_input.x,
            y=np.sin(dc_input.x)
        )
        return output
    ```

    ## equivalent to: decorated_sine = flojoy(SINE)
    ```
    pj_ids = [123, 456]
    logger.debug(SINE(previous_job_ids = pj_ids, mock = True))
    ```
    """

    def decorator(func: Callable[..., Optional[DataContainer | dict[str, Any]]]):
        @wraps(func)
        def wrapper(
            node_id: str,
            job_id: str,
            jobset_id: str,
            previous_jobs: list[dict[str, str]] = [],
            ctrls: dict[str, Any] | None = None,
        ):
            try:
                FN = func.__name__
                sys_status = "🏃‍♀️ Running python job: " + FN
                send_to_socket(
                    json.dumps(
                        {
                            "SYSTEM_STATUS": sys_status,
                            "jobsetId": jobset_id,
                            "RUNNING_NODE": node_id,
                        }
                    )
                )
                logger.debug("previous jobs:", previous_jobs)
                # Get command parameters set by the user through the control panel
                func_params: dict[str, Any] = {}
                if ctrls is not None:
                    for _, input in ctrls.items():
                        param = input["param"]
                        value = input["value"]
                        func_params[param] = format_param_value(value, input["type"])
                func_params["type"] = "default"

                logger.debug(
                    "executing node_id:",
                    node_id,
                    "previous_jobs:",
                    previous_jobs,
                )
                dict_inputs = fetch_inputs(previous_jobs)

                # constructing the inputs
                logger.debug(f"constructing inputs for {func.__name__}")
                args: dict[str, Any] = {}
                sig = signature(func)

                args = {**args, **dict_inputs}

                for param, value in func_params.items():
                    if param in sig.parameters:
                        args[param] = value
                if inject_node_metadata:
                    args["default_params"] = DefaultParams(
                        job_id=job_id,
                        node_id=node_id,
                        jobset_id=jobset_id,
                        node_type="default",
                    )

                logger.debug(node_id, " params: ", args.keys())

                # check if node has an init container and if so, inject it
                if NodeInitService().has_init_store(node_id):
                    args["init_container"] = NodeInitService().get_init_store(node_id)

                ##########################
                # calling the node function
                ##########################
                dc_obj = func(**args)  # DataContainer object from node
                ##########################
                # end calling the node function
                ##########################

                # some special nodes like LOOP return dict instead of `DataContainer`
                if isinstance(dc_obj, DataContainer):
                    dc_obj.validate()  # Validate returned DataContainer object
                elif dc_obj is not None:
                    for value in dc_obj.values():
                        if isinstance(value, DataContainer):
                            value.validate()
                # Response object to send to FE
                result = get_frontend_res_obj_from_result(dc_obj)

                JobService().post_job_result(
                    job_id, dc_obj
                )  # post result to the job service before sending result to socket

                send_to_socket(
                    json.dumps(
                        {
                            "NODE_RESULTS": {
                                "cmd": FN,
                                "id": node_id,
                                "result": result,
                            },
                            "jobsetId": jobset_id,
                        },
                        cls=PlotlyJSONEncoder,
                    )
                )

                if func.__name__ == "END":
                    send_to_socket(
                        json.dumps(
                            {
                                "SYSTEM_STATUS": "🤙 python script run successful",
                                "RUNNING_NODE": "",
                                "jobsetId": jobset_id,
                            }
                        )
                    )
                return dc_obj
            except Exception as e:
                send_to_socket(
                    json.dumps(
                        {
                            "SYSTEM_STATUS": f"Failed to run: {func.__name__}",
                            "FAILED_NODES": {node_id: str(e)},
                            "jobsetId": jobset_id,
                        }
                    )
                )
                logger.debug("error occured while running the node")
                logger.debug(traceback.format_exc())
                raise e

        return wrapper

    if original_function:
        return decorator(original_function)

    return decorator



from typing import Optional, Union, Any, Literal, TypedDict
from pathlib import Path
from .data_container import *
from .flojoy_python import *
from .job_result_builder import *
from .flojoy_instruction import *
from .plotly_utils import *
from .module_scraper import *
from .job_result_utils import *
from .utils import *
from .parameter_types import *
from .small_memory import *
from .flojoy_node_venv import *
from .job_service import *
from .node_init import *
from .node_preflight import *
from .data_container import *
from .config import *
from .flojoy_cloud import *

def hf_hub_download(
    repo_id: str,
    filename: str,
    *,
    subfolder: Optional[str] = None,
    repo_type: Optional[str] = None,
    revision: Optional[str] = None,
    library_name: Optional[str] = None,
    library_version: Optional[str] = None,
    cache_dir: Union[str, Path, None] = None,
    local_dir: Union[str, Path, None] = None,
    local_dir_use_symlinks: Union[bool, Literal["auto"]] = "auto",
    user_agent: Union[dict, str, None] = None,
    force_download: bool = False,
    force_filename: Optional[str] = None,
    proxies: Optional[dict] = None,
    etag_timeout: float = 10,
    resume_download: bool = False,
    token: Union[bool, str, None] = None,
    local_files_only: bool = False,
    legacy_cache_layout: bool = False,
) -> str:
    """Download a given file if it's not already present in the local cache.

    The new cache file layout looks like this:
    - The cache directory contains one subfolder per repo_id (namespaced by repo type)
    - inside each repo folder:
        - refs is a list of the latest known revision => commit_hash pairs
        - blobs contains the actual file blobs (identified by their git-sha or sha256, depending on
          whether they're LFS files or not)
        - snapshots contains one subfolder per commit, each "commit" contains the subset of the files
          that have been resolved at that particular commit. Each filename is a symlink to the blob
          at that particular commit.

    If `local_dir` is provided, the file structure from the repo will be replicated in this location. You can configure
    how you want to move those files:
      - If `local_dir_use_symlinks="auto"` (default), files are downloaded and stored in the cache directory as blob
        files. Small files (<5MB) are duplicated in `local_dir` while a symlink is created for bigger files. The goal
        is to be able to manually edit and save small files without corrupting the cache while saving disk space for
        binary files. The 5MB threshold can be configured with the `HF_HUB_LOCAL_DIR_AUTO_SYMLINK_THRESHOLD`
        environment variable.
      - If `local_dir_use_symlinks=True`, files are downloaded, stored in the cache directory and symlinked in `local_dir`.
        This is optimal in term of disk usage but files must not be manually edited.
      - If `local_dir_use_symlinks=False` and the blob files exist in the cache directory, they are duplicated in the
        local dir. This means disk usage is not optimized.
      - Finally, if `local_dir_use_symlinks=False` and the blob files do not exist in the cache directory, then the
        files are downloaded and directly placed under `local_dir`. This means if you need to download them again later,
        they will be re-downloaded entirely.

    ```
    [  96]  .
    └── [ 160]  models--julien-c--EsperBERTo-small
        ├── [ 160]  blobs
        │   ├── [321M]  403450e234d65943a7dcf7e05a771ce3c92faa84dd07db4ac20f592037a1e4bd
        │   ├── [ 398]  7cb18dc9bafbfcf74629a4b760af1b160957a83e
        │   └── [1.4K]  d7edf6bd2a681fb0175f7735299831ee1b22b812
        ├── [  96]  refs
        │   └── [  40]  main
        └── [ 128]  snapshots
            ├── [ 128]  2439f60ef33a0d46d85da5001d52aeda5b00ce9f
            │   ├── [  52]  README.md -> ../../blobs/d7edf6bd2a681fb0175f7735299831ee1b22b812
            │   └── [  76]  pytorch_model.bin -> ../../blobs/403450e234d65943a7dcf7e05a771ce3c92faa84dd07db4ac20f592037a1e4bd
            └── [ 128]  bbc77c8132af1cc5cf678da3f1ddf2de43606d48
                ├── [  52]  README.md -> ../../blobs/7cb18dc9bafbfcf74629a4b760af1b160957a83e
                └── [  76]  pytorch_model.bin -> ../../blobs/403450e234d65943a7dcf7e05a771ce3c92faa84dd07db4ac20f592037a1e4bd
    ```

    Args:
        repo_id (`str`):
            A user or an organization name and a repo name separated by a `/`.
        filename (`str`):
            The name of the file in the repo.
        subfolder (`str`, *optional*):
            An optional value corresponding to a folder inside the model repo.
        repo_type (`str`, *optional*):
            Set to `"dataset"` or `"space"` if downloading from a dataset or space,
            `None` or `"model"` if downloading from a model. Default is `None`.
        revision (`str`, *optional*):
            An optional Git revision id which can be a branch name, a tag, or a
            commit hash.
        library_name (`str`, *optional*):
            The name of the library to which the object corresponds.
        library_version (`str`, *optional*):
            The version of the library.
        cache_dir (`str`, `Path`, *optional*):
            Path to the folder where cached files are stored.
        local_dir (`str` or `Path`, *optional*):
            If provided, the downloaded file will be placed under this directory, either as a symlink (default) or
            a regular file (see description for more details).
        local_dir_use_symlinks (`"auto"` or `bool`, defaults to `"auto"`):
            To be used with `local_dir`. If set to "auto", the cache directory will be used and the file will be either
            duplicated or symlinked to the local directory depending on its size. It set to `True`, a symlink will be
            created, no matter the file size. If set to `False`, the file will either be duplicated from cache (if
            already exists) or downloaded from the Hub and not cached. See description for more details.
        user_agent (`dict`, `str`, *optional*):
            The user-agent info in the form of a dictionary or a string.
        force_download (`bool`, *optional*, defaults to `False`):
            Whether the file should be downloaded even if it already exists in
            the local cache.
        proxies (`dict`, *optional*):
            Dictionary mapping protocol to the URL of the proxy passed to
            `requests.request`.
        etag_timeout (`float`, *optional*, defaults to `10`):
            When fetching ETag, how many seconds to wait for the server to send
            data before giving up which is passed to `requests.request`.
        resume_download (`bool`, *optional*, defaults to `False`):
            If `True`, resume a previously interrupted download.
        token (`str`, `bool`, *optional*):
            A token to be used for the download.
                - If `True`, the token is read from the HuggingFace config
                  folder.
                - If a string, it's used as the authentication token.
        local_files_only (`bool`, *optional*, defaults to `False`):
            If `True`, avoid downloading the file and return the path to the
            local cached file if it exists.
        legacy_cache_layout (`bool`, *optional*, defaults to `False`):
            If `True`, uses the legacy file cache layout i.e. just call [`hf_hub_url`]
            then `cached_download`. This is deprecated as the new cache layout is
            more powerful.

    Returns:
        Local path (string) of file or if networking is off, last version of
        file cached on disk.

    <Tip>

    Raises the following errors:

        - [`EnvironmentError`](https://docs.python.org/3/library/exceptions.html#EnvironmentError)
          if `token=True` and the token cannot be found.
        - [`OSError`](https://docs.python.org/3/library/exceptions.html#OSError)
          if ETag cannot be determined.
        - [`ValueError`](https://docs.python.org/3/library/exceptions.html#ValueError)
          if some parameter value is invalid
        - [`~utils.RepositoryNotFoundError`]
          If the repository to download from cannot be found. This may be because it doesn't exist,
          or because it is set to `private` and you do not have access.
        - [`~utils.RevisionNotFoundError`]
          If the revision to download from cannot be found.
        - [`~utils.EntryNotFoundError`]
          If the file to download cannot be found.
        - [`~utils.LocalEntryNotFoundError`]
          If network is disabled or unavailable and file is not found in cache.

    </Tip>
    """
    ...

def snapshot_download(
    repo_id: str,
    *,
    revision: Optional[str] = None,
    repo_type: Optional[str] = None,
    cache_dir: Union[str, Path, None] = None,
    local_dir: Union[str, Path, None] = None,
    local_dir_use_symlinks: Union[bool, Literal["auto"]] = "auto",
    library_name: Optional[str] = None,
    library_version: Optional[str] = None,
    user_agent: Optional[Union[dict, str]] = None,
    proxies: Optional[dict] = None,
    etag_timeout: float = 10,
    resume_download: bool = False,
    force_download: bool = False,
    token: Optional[Union[bool, str]] = None,
    local_files_only: bool = False,
    allow_patterns: Optional[Union[list[str], str]] = None,
    ignore_patterns: Optional[Union[list[str], str]] = None,
    max_workers: int = 8,
    tqdm_class: Optional[Any] = None,
) -> str:
    """Download repo files.

    Download a whole snapshot of a repo's files at the specified revision. This is useful when you want all files from
    a repo, because you don't know which ones you will need a priori. All files are nested inside a folder in order
    to keep their actual filename relative to that folder. You can also filter which files to download using
    `allow_patterns` and `ignore_patterns`.

    If `local_dir` is provided, the file structure from the repo will be replicated in this location. You can configure
    how you want to move those files:
      - If `local_dir_use_symlinks="auto"` (default), files are downloaded and stored in the cache directory as blob
        files. Small files (<5MB) are duplicated in `local_dir` while a symlink is created for bigger files. The goal
        is to be able to manually edit and save small files without corrupting the cache while saving disk space for
        binary files. The 5MB threshold can be configured with the `HF_HUB_LOCAL_DIR_AUTO_SYMLINK_THRESHOLD`
        environment variable.
      - If `local_dir_use_symlinks=True`, files are downloaded, stored in the cache directory and symlinked in `local_dir`.
        This is optimal in term of disk usage but files must not be manually edited.
      - If `local_dir_use_symlinks=False` and the blob files exist in the cache directory, they are duplicated in the
        local dir. This means disk usage is not optimized.
      - Finally, if `local_dir_use_symlinks=False` and the blob files do not exist in the cache directory, then the
        files are downloaded and directly placed under `local_dir`. This means if you need to download them again later,
        they will be re-downloaded entirely.

    An alternative would be to clone the repo but this requires git and git-lfs to be installed and properly
    configured. It is also not possible to filter which files to download when cloning a repository using git.

    Args:
        repo_id (`str`):
            A user or an organization name and a repo name separated by a `/`.
        revision (`str`, *optional*):
            An optional Git revision id which can be a branch name, a tag, or a
            commit hash.
        repo_type (`str`, *optional*):
            Set to `"dataset"` or `"space"` if downloading from a dataset or space,
            `None` or `"model"` if downloading from a model. Default is `None`.
        cache_dir (`str`, `Path`, *optional*):
            Path to the folder where cached files are stored.
        local_dir (`str` or `Path`, *optional*:
            If provided, the downloaded files will be placed under this directory, either as symlinks (default) or
            regular files (see description for more details).
        local_dir_use_symlinks (`"auto"` or `bool`, defaults to `"auto"`):
            To be used with `local_dir`. If set to "auto", the cache directory will be used and the file will be either
            duplicated or symlinked to the local directory depending on its size. It set to `True`, a symlink will be
            created, no matter the file size. If set to `False`, the file will either be duplicated from cache (if
            already exists) or downloaded from the Hub and not cached. See description for more details.
        library_name (`str`, *optional*):
            The name of the library to which the object corresponds.
        library_version (`str`, *optional*):
            The version of the library.
        user_agent (`str`, `dict`, *optional*):
            The user-agent info in the form of a dictionary or a string.
        proxies (`dict`, *optional*):
            Dictionary mapping protocol to the URL of the proxy passed to
            `requests.request`.
        etag_timeout (`float`, *optional*, defaults to `10`):
            When fetching ETag, how many seconds to wait for the server to send
            data before giving up which is passed to `requests.request`.
        resume_download (`bool`, *optional*, defaults to `False):
            If `True`, resume a previously interrupted download.
        force_download (`bool`, *optional*, defaults to `False`):
            Whether the file should be downloaded even if it already exists in the local cache.
        token (`str`, `bool`, *optional*):
            A token to be used for the download.
                - If `True`, the token is read from the HuggingFace config
                  folder.
                - If a string, it's used as the authentication token.
        local_files_only (`bool`, *optional*, defaults to `False`):
            If `True`, avoid downloading the file and return the path to the
            local cached file if it exists.
        allow_patterns (`List[str]` or `str`, *optional*):
            If provided, only files matching at least one pattern are downloaded.
        ignore_patterns (`List[str]` or `str`, *optional*):
            If provided, files matching any of the patterns are not downloaded.
        max_workers (`int`, *optional*):
            Number of concurrent threads to download files (1 thread = 1 file download).
            Defaults to 8.
        tqdm_class (`tqdm`, *optional*):
            If provided, overwrites the default behavior for the progress bar. Passed
            argument must inherit from `tqdm.auto.tqdm` or at least mimic its behavior.
            Note that the `tqdm_class` is not passed to each individual download.
            Defaults to the custom HF progress bar that can be disabled by setting
            `HF_HUB_DISABLE_PROGRESS_BARS` environment variable.

    Returns:
        Local folder path (string) of repo snapshot

    <Tip>

    Raises the following errors:

    - [`EnvironmentError`](https://docs.python.org/3/library/exceptions.html#EnvironmentError)
      if `token=True` and the token cannot be found.
    - [`OSError`](https://docs.python.org/3/library/exceptions.html#OSError) if
      ETag cannot be determined.
    - [`ValueError`](https://docs.python.org/3/library/exceptions.html#ValueError)
      if some parameter value is invalid

    </Tip>
    """
    ...

def flojoy(
    original_function: Callable[..., DataContainer | dict[str, Any] | TypedDict | None]
    | None = None,
    *,
    node_type: Optional[str] = None,
    deps: Optional[dict[str, str]] = None,
    inject_node_metadata: bool = False,
) -> Callable[..., DataContainer | dict[str, Any] | None]: ...



import logging

LOGGER_NAME = "flojoy"


class FlojoyConfig:
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = FlojoyConfig()
        return cls._instance

    def __init__(self):
        self.is_offline = False


logger = logging.getLogger(LOGGER_NAME)



class FLOJOY_INSTRUCTION:
    FLOW_TO_DIRECTIONS = "__flow_to_directions__"
    FLOW_TO_NODES = "__flow_to_nodes__"
    RESULT_FIELD = "__result_field__"
    DATACONTAINER_FIELD = "__data_container_field__"



from .data_container import DataContainer, OrderedPair
import numpy as np
from typing import Union
from .flojoy_instruction import FLOJOY_INSTRUCTION


class JobResultBuilder:
    instructions: dict[str, Union[str, list[str]]] | None = None

    def __init__(self) -> None:
        self.data = None

    def _add_instructions(self, instruction: dict[str, Union[str, list[str]]]):
        self.instructions = self.instructions if self.instructions is not None else {}
        self.instructions = {
            **self.instructions,
            **instruction,
        }

    def from_inputs(self, inputs: list[DataContainer]):
        # if no inputs were provided, construct fake output
        if len(inputs) == 0:
            self.data = None
        else:
            self.data = inputs[0]

        return self

    def from_data(self, data: DataContainer):
        self.data = data
        return self

    def flow_to_nodes(self, nodes: list[str]):
        if nodes.__len__() > 0:
            self._add_instructions({FLOJOY_INSTRUCTION.FLOW_TO_NODES: nodes})
        return self

    def flow_to_directions(self, directions: list[str]):
        if directions.__len__() > 0:
            self._add_instructions({FLOJOY_INSTRUCTION.FLOW_TO_DIRECTIONS: directions})
        return self

    def flow_by_flag(
        self, flag: bool, true_direction: list[str], false_direction: list[str]
    ):
        self._add_instructions(
            {
                FLOJOY_INSTRUCTION.FLOW_TO_DIRECTIONS: true_direction
                if flag
                else false_direction
            }
        )
        return self

    def build(self):
        result = self.data
        if self.instructions:
            result = {
                # instructions to job scheduler (watch.py)
                **self.instructions,
                # instruction to fetch_input method in flojoy.py
                FLOJOY_INSTRUCTION.RESULT_FIELD: "data",
                "data": result,
            }
        return result

    def get_default_data(self) -> DataContainer:
        x = np.arange(0, 1000, 1)  # type:ignore
        y = np.ones_like(x)
        return OrderedPair(x=x, y=y)



"""
This module provides a decorator that allows a function to be executed in a virtual environment.
The decorator will create a virtual environment if it does not exist, and install the pip dependencies
specified in the decorator arguments. The decorator will also install the pip dependencies if the
virtual environment exists but does not contain the pip dependencies.

Example usage:

```python
from flojoy import flojoy, run_in_venv

@flojoy
@run_in_venv(pip_dependencies=["torch==2.0.1", "torchvision==0.15.2"])
def TORCH_NODE(default: Matrix) -> Matrix:
    import torch
    import torchvision
    # Do stuff with torch
    ...
    return Matrix(...)

"""
from collections.abc import Callable, Iterable, Mapping
from concurrent.futures import Future, ThreadPoolExecutor
import signal
import threading
from time import sleep
from typing import Any, Callable, Optional

import hashlib
from contextlib import ExitStack, contextmanager
import importlib.metadata
import inspect
import logging
import multiprocessing
import multiprocessing.connection
import os
import shutil
import re
import subprocess
import sys
import traceback
import venv
from functools import wraps
import cloudpickle

from .utils import FLOJOY_CACHE_DIR
from .logging import LogPipe, LogPipeMode, StreamEnum

__all__ = ["run_in_venv"]


def _get_venv_cache_dir():
    return os.path.join(FLOJOY_CACHE_DIR, "flojoy_node_venv")


def _get_venv_syspath(venv_executable: os.PathLike) -> list[str]:
    """Get the sys.path of the virtual environment."""
    command = [venv_executable, "-c", "import sys\nprint(sys.path)"]
    cmd_output = subprocess.run(command, check=True, capture_output=True, text=True)
    return eval(cmd_output.stdout)


@contextmanager
def swap_sys_path(
    venv_executable: os.PathLike, extra_sys_path: list[str] | None = None
):
    """Temporarily swap the sys.path of the child process with the sys.path of the parent process."""
    old_path = sys.path
    try:
        new_path = _get_venv_syspath(venv_executable)
        extra_sys_path = [] if extra_sys_path is None else extra_sys_path
        sys.path = new_path + extra_sys_path
        yield
    finally:
        sys.path = old_path


def _get_venv_executable_path(venv_path: os.PathLike | str) -> os.PathLike | str:
    """Get the path to the python executable of the virtual environment."""
    if sys.platform == "win32":
        return os.path.realpath(os.path.join(venv_path, "Scripts", "python.exe"))
    else:
        return os.path.realpath(os.path.join(venv_path, "bin", "python"))


def _bootstrap_venv(
    venv_path: os.PathLike,
    pip_dependencies: list[str],
    logger: logging.Logger,
    verbose: bool = False,
):
    venv_executable = _get_venv_executable_path(venv_path)
    if not os.path.exists(venv_executable):
        if os.path.exists(venv_path):
            logger.warning(
                f"For some reason, the virtual environment at {venv_path} does not contain a python executable. Deleting the virtual environment and creating a new one..."
            )
            shutil.rmtree(venv_path, ignore_errors=True)
        logger.info(f"Creating new virtual environment at {venv_path}...")
        venv.create(venv_path, with_pip=True)
    # Deref the symlink -> Edge case with windows
    venv_executable = os.path.realpath(venv_executable)
    command = [venv_executable, "-m", "pip", "install"]
    if not verbose:
        command += ["-q", "-q"]
    command += list(pip_dependencies)
    with ExitStack() as stack:
        logpipe_stderr = stack.enter_context(
            LogPipe(
                logger,
                log_level=logging.ERROR,
                mode=LogPipeMode.SUBPROCESS,
                buffer_logs=True,
            )
        )
        logpipe_stdout = stack.enter_context(
            LogPipe(
                logger,
                log_level=logging.INFO,
                mode=LogPipeMode.SUBPROCESS,
                buffer_logs=True,
            )
        )
        proc = subprocess.Popen(
            command,
            stdout=logpipe_stdout.get_pipe_writer(),
            stderr=logpipe_stderr.get_pipe_writer(),
        )
        # Poll the process until it finishes, while occasionally checking if there was a failure
        # in the global cancel event
        while proc.poll() is None:
            if PipInstallThread._cancel_all_threads.is_set():
                proc.terminate()
                logger.error(
                    f"Another thread has failed its pip install step, terminating pip install process for {venv_path}..."
                )
                break
            sleep(0.1)

    venv_complete_path = os.path.realpath(os.path.join(venv_path, ".venv_is_complete"))

    # The process was terminated due to the _cancel_all_threads event
    if proc.returncode is None:
        return

    if proc.returncode != 0:
        if os.path.exists(venv_complete_path):
            os.remove(venv_complete_path)
        shutil.rmtree(venv_path, ignore_errors=True)
        logger.error(
            f"Failed to install pip dependencies into virtual environment from the provided list: {pip_dependencies}. The virtual environment under {venv_path} has been deleted."
        )
        raise subprocess.CalledProcessError(
            proc.returncode,
            command,
            output=logpipe_stdout.log_buffer.getvalue().encode("utf-8"),
            stderr=logpipe_stdout.log_buffer.getvalue().encode("utf-8"),
        )

    # Create a file to mark the virtual environment as complete
    with open(os.path.join(venv_path, ".venv_is_complete"), "w") as f:
        f.write("")


class PipInstallThread(threading.Thread):
    _bounded_semaphore = threading.BoundedSemaphore(4)
    _cancel_all_threads = threading.Event()
    _threads = dict()
    _exceptions = dict()

    def __init__(
        self,
        group: None = None,
        target: Callable[..., object] | None = None,
        name: str | None = None,
        args: Iterable[Any] = ...,
        kwargs: Mapping[str, Any] | None = None,
    ) -> None:
        super().__init__(group, target, name, args, kwargs, daemon=False)
        self.target = target
        self.args = args or tuple()
        self.kwargs = kwargs or {}
        PipInstallThread._threads.update({self.name: self})
        PipInstallThread._exceptions.update({self.name: None})

    def run(self):
        with PipInstallThread._bounded_semaphore:
            # Just exit if all was cancelled
            if PipInstallThread._cancel_all_threads.is_set():
                return
            try:
                self.target(*self.args, **self.kwargs)
            except Exception as e:
                PipInstallThread._exceptions.update({self.name: e})
                raise e

    @staticmethod
    def terminate_all():
        PipInstallThread._cancel_all_threads.set()
        while all([thread.is_alive() for thread in PipInstallThread._threads.values()]):
            sleep(0.1)


class PickleableFunctionWithPipeIO:
    """A wrapper for a function that can be pickled and executed in a child process."""

    def __init__(
        self,
        func: Callable,
        child_conn: multiprocessing.connection.Connection,
        venv_executable: str,
    ):
        self._func_serialized = cloudpickle.dumps(func)
        func_module_path = os.path.dirname(os.path.realpath(inspect.getabsfile(func)))
        # Check that the function is in a directory indeed
        self._extra_sys_path = (
            [func_module_path] if os.path.isdir(func_module_path) else None
        )
        self._child_conn = child_conn
        self._venv_executable = venv_executable

    def __call__(self, *args_serialized, **kwargs_serialized):
        with swap_sys_path(
            venv_executable=self._venv_executable, extra_sys_path=self._extra_sys_path
        ):
            try:
                fn = cloudpickle.loads(self._func_serialized)
                args = [cloudpickle.loads(arg) for arg in args_serialized]
                kwargs = {
                    key: cloudpickle.loads(value)
                    for key, value in kwargs_serialized.items()
                }
                # Capture logs here too
                output = fn(*args, **kwargs)
                serialized_result = cloudpickle.dumps(output)
            except Exception as e:
                # Not all exceptions are expected to be picklable
                # so we clone their traceback and send our own custom type of exception
                exc = ChildProcessError(
                    f"Child process failed with an exception of type {type(e)}."
                ).with_traceback(e.__traceback__)
                serialized_result = cloudpickle.dumps(
                    (exc, traceback.format_exception(type(e), e, e.__traceback__))
                )
        self._child_conn.send_bytes(serialized_result)


def run_in_venv(pip_dependencies: list[str], verbose: bool = False):
    """A decorator that allows a function to be executed in a virtual environment.

    Args:
        pip_dependencies (list[str]): A list of pip dependencies to install into the virtual environment.
        verbose (bool): Whether to print the pip install output. Defaults to False.

    Example usage:
    ```python
    from flojoy import flojoy, run_in_venv

    @flojoy
    @run_in_venv(pip_dependencies=["torch==2.0.1", "torchvision==0.15.2"])
    def TORCH_NODE(default: Matrix) -> Matrix:
        import torch
        import torchvision
        # Do stuff with torch
        ...
        return Matrix(...)
    """

    def decorator(func, *, pip_dependencies=pip_dependencies, verbose=verbose):
        # Return the function as-is if we are not in the main process
        # This is due to the fact that the decorator is called twice
        # once in the main process and once in the child process
        # when unpickling the function
        if multiprocessing.current_process().name.startswith("run_in_venv"):
            return func

        # Pre-pend flojoy and cloudpickle as mandatory pip dependencies
        packages_dict = {
            package.name: package.version
            for package in importlib.metadata.distributions()
        }
        pip_dependencies = sorted(
            [
                f"flojoy=={packages_dict['flojoy']}",
                f"cloudpickle=={packages_dict['cloudpickle']}",
            ]
            + pip_dependencies
        )
        # Get the root directory for the virtual environments
        venv_cache_dir = _get_venv_cache_dir()
        os.makedirs(venv_cache_dir, exist_ok=True)
        venv_cache_dir = os.path.realpath(venv_cache_dir)
        # Generate a path-safe hash of the pip dependencies
        # this prevents the duplication of virtual environments
        pip_dependencies_hash = hashlib.md5(
            "".join(sorted(pip_dependencies)).encode()
        ).hexdigest()[:8]
        venv_path = os.path.join(venv_cache_dir, f"{pip_dependencies_hash}")
        venv_executable = _get_venv_executable_path(venv_path)
        logger = logging.getLogger(func.__name__)
        if verbose:
            logger.setLevel(logging.INFO)

        thread = PipInstallThread(
            target=_bootstrap_venv,
            args=(venv_path, pip_dependencies, logger, verbose),
        )
        thread.start()

        @wraps(func)
        def wrapper(*args, **kwargs):
            # Wait for the pip install to finish
            logger.info(
                f"Waiting for pip install to finish for virtual environment of {func.__name__} at  {venv_path}..."
            )
            thread.join()
            venv_executable = os.path.realpath(venv_executable)
            # Check if the thread threw an exception
            if PipInstallThread._exceptions[thread.name] is not None:
                # Clean up the other threads (and the processes they spawned)
                PipInstallThread.terminate_all()
                # Re-raise from the main thread
                raise PipInstallThread._exceptions[thread.name]
            logger.info(
                f"Pip install complete. Spawning process for function {func.__name__}..."
            )
            # Generate a new multiprocessing context for the parent process in "spawn" mode
            parent_mp_context = multiprocessing.get_context("spawn")
            parent_conn, child_conn = parent_mp_context.Pipe()
            # Create a new multiprocessing context for the child process in "spawn" mode
            # while setting its executable to the virtual environment python executable
            child_mp_context = multiprocessing.get_context("spawn")
            child_mp_context.set_executable(venv_executable)
            with ExitStack() as stack:
                log_pipe_stdout = stack.enter_context(
                    LogPipe(
                        logger=logger, log_level=logging.INFO, mode=LogPipeMode.MP_SPAWN
                    )
                )
                log_pipe_stderr = stack.enter_context(
                    LogPipe(
                        logger=logger,
                        log_level=logging.ERROR,
                        mode=LogPipeMode.MP_SPAWN,
                    )
                )
                # Serialize function arguments using cloudpickle
                pickleable_func_with_pipe = PickleableFunctionWithPipeIO(
                    func=func,
                    child_conn=child_conn,
                    venv_executable=venv_executable,
                )
                # Wrap the function with a decorator that redirects stdout and stderr to the log pipes
                mp_func = LogPipe.wrap_and_redirect_stream(
                    pickleable_func_with_pipe,
                    StreamEnum.STDOUT,
                    log_pipe_stdout.get_pipe_writer(),
                )
                mp_func = LogPipe.wrap_and_redirect_stream(
                    mp_func, StreamEnum.STDERR, log_pipe_stderr.get_pipe_writer()
                )
                # Resolve the function arguments using inspect
                # this is needed to avoid pickling issues
                kwargs_serialized = {
                    key: cloudpickle.dumps(value)
                    for key, value in inspect.getcallargs(func, *args, **kwargs).items()
                }
                # Create a new process that will run the Python code
                # Append a name and a random suffix
                process = child_mp_context.Process(
                    name=f"run_in_venv_{os.urandom(4).hex()}",
                    target=mp_func,
                    kwargs=kwargs_serialized,
                )
                # Start the process
                process.start()
                # Fetch result from the child process
                serialized_result = parent_conn.recv_bytes()
                # Wait for the process to finish
                process.join()
            # Check if the process sent an exception with a traceback
            result = cloudpickle.loads(serialized_result)
            if isinstance(result, tuple) and isinstance(result[0], Exception):
                # Fetch exception and formatted traceback (list[str])
                exception, tcb = result
                # Reraise an exception with the same class
                logger.error(
                    f"Error in child process with the following traceback:\n{''.join(tcb)}"
                )
                raise exception
            return result

        return wrapper

    return decorator



import inspect


class FlojoyWrapper:
    CUSTOM_DOC_ADDITION = "-." * 36 + "\n\t"
    CUSTOM_DOC_ADDITION += (
        """The parameters of the function in this Flojoy wrapper are given below."""
    )
    CUSTOM_DOC_ADDITION += "\n\t" + "-." * 36 + "\n"
    INPUT_DEFAULT_DTYPE = "np.ndarray"
    FORBIDDEN_OPTIONAL_ARGS = ["x", "data", "kwargs", "comparator", "a", "A"]
    FORBIDDEN_TYPES = [
        "tuple",
        "array-like",
        "array_like",
        "function",
        "callable",
        "sequence",
    ]

    def __init__(self, func, parameters, module, argument_names):
        # we'll use parameters to get the actual default values of the arguments
        # for generating the manifest
        self.name = func.__name__
        self.doc = func.__doc__
        self.arguments = argument_names
        self.optional_argument_dict = parameters
        self.module = module
        try:
            self.first_argument = argument_names[0]
            self.parameters = {
                self.first_argument: {
                    "dtype": self.INPUT_DEFAULT_DTYPE,
                    "optional": False,
                }
            }
        except IndexError:  # this means argument_names is none
            self.first_argument = None
            self.parameters = {}
        self.data = f"from flojoy import DataContainer, flojoy\n"
        self.data += f"import {module.__name__}\n\n"
        self.data += f"@flojoy\ndef {self.name.upper()}(dc, params):\n\t"

        self.manifest = "COMMAND:\n\t- "

        self.process_docstring()

    def process_docstring(self):
        for idl, line in enumerate(self.doc.split("\n")):
            if "Returns" in line:
                break
        self.doc = "\n" + (
            "\n".join(
                [
                    ("\t" if (":" in line or "----------" in line) else "\t\t")
                    + line.lstrip(" ")
                    for line in self.doc.split("\n")[:idl]
                ]
            ).replace("\tParameters", self.CUSTOM_DOC_ADDITION + "\n" + "\tParameters")
        )

    def write_manifest(self, mtype):
        self.manifest += "name: " + self.name.lower().capitalize() + "\n"
        self.manifest += "\t\tkey: " + self.name.upper() + "\n"
        self.manifest += "\t\ttype: " + mtype + "\n"
        parameter_header = "\t\tparameters: \n\t\t"
        parameter_string = ""
        if self.parameters.keys() is not []:
            for param in self.parameters.keys():
                if param not in self.FORBIDDEN_OPTIONAL_ARGS:
                    try:
                        def_val = str(self.optional_argument_dict[param])
                    except:
                        def_val = "None"
                    dtype = str(self.parameters[param]["dtype"])
                    # sometimes we get 'None or float' ...
                    if dtype == "str":
                        dtype = "string"
                    elif dtype == "None or float":
                        dtype = "float"
                    elif dtype == "bool":
                        dtype = "boolean"
                    parameter_string += (
                        f"\t{param}: \n"
                        + "\t\t\t\ttype: "
                        + dtype
                        + "\n\t\t\t\tdefault: "
                        + ("" if def_val == "None" else def_val)
                        + " \n\t\t"
                    )
            if parameter_string != "":
                self.manifest += parameter_header + parameter_string

            self.manifest += "\t\t\n"
        self.manifest = self.manifest.replace("\t", "  ")

    def write_wrapper(self, mtype):
        if "callable" in self.doc:
            # logger.debug(
            # 	"#CANNOT PROCESS ",
            # 	self.name,
            # 	"since",
            # 	"callable",
            # 	"is not allowed ...",
            # )
            self.data = ""
            self.manifest = ""
            return
        self.data += """'''"""
        self.data += self.doc
        self.data += "\t" + """'''""" + "\n"
        self.data += f"\treturn DataContainer("
        if self.module.__name__ != "numpy.matlib":
            self.data += "\n\t\tx=dc[0].y,\n\t\t"
            self.data += f"y={self.module.__name__}.{self.name}(\n\t\t\t" + (
                f"{self.first_argument}=dc[0].y,\n\t\t\t"
                if self.first_argument is not None
                else ""
            )
        else:
            self.data += f"\n\t\tm={self.module.__name__}.{self.name}(\n\t\t\t" + (
                f"{self.first_argument}=dc[0].y,\n\t\t\t"
                if self.first_argument is not None
                else ""
            )
        for idk, arg in enumerate(self.arguments):
            if arg not in self.FORBIDDEN_OPTIONAL_ARGS:
                dtype = ""
                try:  # try to read it from the inspect dictionary
                    dtype = type(self.optional_argument_dict[arg]).__name__
                except:
                    pass
                if dtype == "" or dtype == "NoneType":
                    for idl, line in enumerate(self.doc.split("\n")):
                        if f"{arg} :" in line:
                            dtype = (
                                line.split(":")[1]
                                .split(",")[0]
                                .strip()
                                .replace("}", "")
                                .replace("{", "")
                                .replace("(", "")
                                .replace(")", "")
                            )
                            break
                # now check for not allowed types after identifying them all correctly:
                if dtype in self.FORBIDDEN_TYPES:
                    # logger.debug(
                    #     "#CANNOT PROCESS ",
                    #     self.name,
                    #     "since",
                    #     dtype,
                    #     "is not allowed ...",
                    # )
                    self.data = ""
                    self.manifest = ""
                    return
                if "ndarray" in dtype:
                    dtype = "np.ndarray"
                    self.data = "import numpy as np" + self.data
                self.data += f"{arg}=({dtype}(params['{arg}']) if params['{arg}'] != '' else None),\n\t\t"

                self.parameters[arg] = {"dtype": dtype, "optional": True}
                self.data += "\t" if idk < len(self.arguments) - 1 else ""
        self.data += ")\n\t)\n\n"
        self.write_manifest(mtype)

    def __repr__(self):
        retval = f"#{self.name}, {self.parameters}\n"
        retval += "#" + "-" * 72 + "\n#Wrapper\n" + "#" + "-" * 72 + "\n"
        retval += self.data + "\n"
        retval += "#" + "-" * 72 + "\n#Manifest\n" + "#" + "-" * 72 + "\n"
        retval += "#" + self.manifest.replace("\n", "\n#") + "\n"
        return retval


def scrape_function(func):
    signature = inspect.signature(func)
    param_names = [p.name for p in signature.parameters.values()]
    default_optional_params = {
        k: val.default
        for k, val in signature.parameters.items()
        if val.default != inspect.Parameter.empty
    }
    return func, param_names, default_optional_params


if __name__ == "__main__":
    import os
    from pathlib import Path
    import scipy
    import ast
    import numpy as np

    MODULES_TO_SCRAPE = {
        "scipy": [scipy.signal, scipy.stats],
        "numpy": [np.linalg, np.random.mtrand],
    }

    CWD = os.getcwd()
    MANIFEST_DIR = CWD / Path("MANIFEST")
    MANIFEST_DIR.mkdir(exist_ok=True)

    for module in MODULES_TO_SCRAPE.keys():
        MODULE_DIR = Path(CWD) / Path(f"{module.upper()}")
        MODULE_DIR.mkdir(exist_ok=True)
        for submodule in MODULES_TO_SCRAPE[module]:
            invalids = []
            valids = []
            submodule_name = submodule.__name__.split(".")[-1]
            NODE_DIR = MODULE_DIR / Path(f"{submodule_name.upper()}")
            NODE_DIR.mkdir(exist_ok=True)

            iterable = (
                inspect.getmembers(submodule, inspect.isfunction)
                if submodule_name != "random"
                else inspect.getmembers(submodule)
            )
            for name, func in inspect.getmembers(submodule, inspect.isfunction):
                _, all_arg_names, default_optional_params = scrape_function(func)
                # for the random library we need to check if there are any input arguments at all!
                func_is_valid = False
                if not all_arg_names:
                    func_is_valid = True
                else:
                    if (all_arg_names[0] in ["x", "data", "A", "a"]) and (
                        "y" not in all_arg_names
                        and "plot" not in all_arg_names
                        and "b" not in all_arg_names
                    ):
                        func_is_valid = True
                if func_is_valid:
                    fw = FlojoyWrapper(
                        func, default_optional_params, submodule, all_arg_names
                    )
                    fw.write_wrapper(f"{module.upper()}_{submodule_name.upper()}")
                    if (
                        fw.manifest != ""
                        and fw.data != ""
                        and "NoneType" not in fw.data
                    ):
                        try:
                            valid = ast.parse(fw.data)
                            this_nodes_directory = Path(NODE_DIR / f"{fw.name.upper()}")
                            this_nodes_directory.mkdir(exist_ok=True)
                            with open(
                                this_nodes_directory / f"{fw.name.upper()}.py", "w"
                            ) as fh:
                                fh.write(fw.data)
                            with open(
                                MANIFEST_DIR / f"{fw.name}.manifest.yaml", "w"
                            ) as fh:
                                fh.write(fw.manifest)
                            valids.append(fw.name)
                        except SyntaxError:
                            invalids.append(fw.name)
            with open(NODE_DIR / "__init__.py", "w") as fh:
                functions = "__all__ = ["
                for name in valids:
                    functions += '"' + name.upper() + '", '
                functions = functions[:-2] + "]\n"
                fh.write(functions)
